// Copyright (C) 1996 DIMACS Center, Rutgers, The State University of New Jersey
// Author(s): Jonathan Berry

// This software is copyrighted by the DIMACS Center at Rutgers, The State
// University of New Jersey.  IT IS PROVIDED AS IS, AND THE AUTHORS, DIMACS, AND
// RUTGERS, THE STATE UNIVERSITY OF NEW JERSEY  DISCLAIM
// ALL LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
// DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

// THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
// IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
// NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
// MODIFICATIONS.

// The authors hereby grant permission to use, copy, modify, distribute,
// and license this software and its documentation for any purpose, provided
// that existing copyright notices are retained in all copies and that this
// notice is included verbatim in any distributions. No written agreement,
// license, or royalty fee is required for any of the authorized uses.
// Modifications to this software may be copyrighted by their authors
// and need not follow the licensing terms described here, provided that
// the new terms are clearly indicated on the first page of each file where
// they apply.

// Last File Update: 31-Jul-1996
// 

#ifndef Collection_h
#define Collection_h

#include "fstream.h"
#include <LINK/basic/Iterator.h>
#include <LINK/basic/general.h>
#include <LINK/basic/ElementOps.h> // if DEFINE_TEMPLATES, methods defs are
				   // needlessly included.  Rely on optimizer.
				   // alternative would be new constant:
				   // DEFINE_E_TEMPLATES.

template <class Item> 
class Collection {
public:
    virtual ~Collection() {}
    virtual int		size() const = 0;
    virtual Bool	emptyQ() const = 0;
    virtual Bool	fullQ() const = 0;
    virtual Bool	memberQ(const Item& e) const = 0;
    virtual Bool	sortedQ() const = 0;
    virtual DataType	type() const = 0;

    virtual Item	first() const = 0;
    virtual Item	last() const = 0;

    virtual void	insert(const Item e) = 0;
    virtual void	append(const Item e) = 0;
    virtual void 	remove(const Item e) = 0;
    virtual void	clear() = 0;
    virtual int         iterate(Iterator<Item>& i, Item& e) const = 0;
    virtual int		occurrences(const Item item) const = 0;
    virtual int		rank(Item e) const = 0;
    virtual Item	ref(int k) const = 0;
    virtual Bool	permutationQ() const = 0;
    virtual Bool	subsetQ(const Collection<Item>&) = 0;
    virtual Bool	properSubsetQ(const Collection<Item>&) = 0;

    void		concatenate(const Collection<Item>& c);

    virtual ostream& 	display(ostream& os) const = 0;
    virtual Collection<Item>* 	newEmpty() const =0;

    virtual Bool        operator==(const Collection<Item>& m) const = 0;
    virtual Bool        operator<=(const Collection<Item>& m) const = 0;
    virtual Bool        operator<(const Collection<Item>& m) const = 0;
    virtual Bool        operator>=(const Collection<Item>& m) const = 0;
    virtual Bool        operator>(const Collection<Item>& m) const = 0;
    virtual Bool        operator!=(const Collection<Item>& m) const = 0;

    virtual Bool        operator==(const Container<Item>& m) const = 0;
    virtual Bool        operator<=(const Container<Item>& m) const = 0;
    virtual Bool        operator<(const Container<Item>& m) const = 0;
    virtual Bool        operator>=(const Container<Item>& m) const = 0;
    virtual Bool        operator>(const Container<Item>& m) const = 0;
    virtual Bool        operator!=(const Container<Item>& m) const = 0;

    friend ostream& operator<<(ostream& os, const Collection<Item>& coll);
};


#ifdef DEFINE_Collection_TEMPLATE
#include <LINK/basic/Collection.cpp>
#endif


#endif
